/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2000 Ian Zimmerman <itz@speakeasy.org> */

static const char rcsid[] = "$Id: preadwrite.c,v 1.11 2002/06/07 22:39:05 itz Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>

#include "preadwrite.h"
#include "exc.h"
#include "gim_common.h"

#define GIM_READ_TIMEOUT 3      /* seconds to wait before giving up on a socket read */

static void
writen (int fd, const void* p, size_t n)
{
  size_t nleft;
  ssize_t nwritten;
  const char* cp;

  cp = p;
  nleft = n;
  while (nleft > 0) {
    if ( (nwritten = write(fd, cp, nleft)) < 0 && errno == EINTR) {
      continue;
    } else if (nwritten < 0 && errno != EINTR) {
      exc_eraise2(unix_error, errno);
    } else {
      assert(nwritten > 0);
      nleft = nleft - nwritten;
      cp = cp + nwritten;
    } /*if*/
  } /*while*/
}

static int
readn (int fd, void* p, size_t n)
{
  size_t nleft;
  ssize_t nread;
  char* cp;
  fd_set read_set;
  struct timeval tv;
  int i;

  cp = p;
  nleft = n;
  while (nleft > 0) {
                                /* this is to avoid blocking forever due to losing peer */
    FD_ZERO(&read_set);
    FD_SET(fd, &read_set);
    tv.tv_sec = GIM_READ_TIMEOUT;
    tv.tv_usec = 0;
    if ( (i = select(fd+1, &read_set, 0, 0, &tv)) < 0 && errno == EINTR) {
      continue;
    } else if (i < 0 && errno != EINTR) {
      exc_eraise2(unix_error, errno);
    } else if (i == 0) {
      break;
    } else {
      assert (i > 0 && FD_ISSET(fd, &read_set));
      if ( (nread = read(fd, cp, nleft)) < 0 && errno == EINTR) {
        continue;
      } else if (nread < 0 && errno != EINTR) {
        exc_eraise2(unix_error, errno);
      } else if (nread == 0) {
        exc_eraise2(eof, n);
      } else {
        assert(nread > 0);
        nleft = nleft - nread;
        cp = cp + nread;
      } /*if*/
    } /*if*/
  } /*while*/
  return n - nleft;
}
                                /* protocol write: prefix with length as 2-digit hex # */
void
gim_pwrite (int fd, const char* p, size_t n)
{
  char prefix[3];
  int i;

  if (n > UCHAR_MAX) exc_eraise2(protocol_error, n);
  i = snprintf(prefix, sizeof(prefix), "%02x", n);
  assert(i == 2);
  writen(fd, prefix, 2);
  writen(fd, p, n);
}
                                /* protocol read: get prefix first to learn length */
size_t
gim_pread (int fd, char* p, size_t n)
{
  char prefix[3];
  size_t len;

  if (readn(fd, prefix, 2) < 2) exc_eraise(protocol_error);
  prefix[2] = '\0';
  if (sscanf(prefix, "%x", &len) != 1)  exc_eraise(protocol_error);
  assert(len >= 0 && len <= UCHAR_MAX);
  if (len > n) exc_eraise(buffer_too_small);
  if (readn(fd, p, len) < len) exc_eraise(protocol_error);
  return len;
}
                                /* protocol transact: send request, get reply, scan error code */
int
gim_transact (int fd, const char* request, char** reply, size_t n)
{
  int code;
  char* p;
  size_t len;

  if (request != 0) gim_pwrite(fd, request, strlen(request));
  len = gim_pread(fd, *reply, n);
  (*reply)[len-1] = '\0';
  code = atoi(*reply);
  /* ignore arriving events until we get a reply */
  while (code < 120) {
    len = gim_pread(fd, *reply, n);
    (*reply)[len-1] = '\0';
    code = atoi(*reply);
  } /*while*/
  p = *reply + strcspn(*reply, " \t\r\n");
  p = p + strspn(p, " \t\r\n");
  *reply = p;
  return code;
}
